{ POLYNOME.PAS : Polynomial decomposition with Horner's scheme

  Title   : POLYNOME
  Language: Borland Pascal v4.0 and higher
  Version : 2.2
  Date    : Feb 15, 2000
  Author  : J R Ferguson
  Usage   : Interactive DOS application
  Download: http://hello.to/ferguson
  E-mail  : j.r.ferguson@iname.com

This program and its source may be used and copied freely without charge,
but  only  for non-commercial purposes. The author is not responsible for
any damage or loss of data that may be caused by using it.

To compile this source file, you wil need  some  units  from  the  JRFPAS
Pascal  routine  library by the same author, which can be downloaded from
the Internet address mentioned above.
}

program POLYNOME(input, output);
uses ConLib, Crt;

const
  C_MaxDegree = 40;
  C_NumDigits =  2; { number of digits of C_MaxDegree }
{*} test      = false;

type
  alfa        = array[1..8] of char;
  index       = 0..C_MaxDegree;
  PolynmTyp   = record
                  degree : index;
                  curptr : index;
                  denom  : integer;
                  coeff  : array[index] of integer
                end;
var
  polynomial  : PolynmTyp;
{*} debug     : boolean;

procedure Cr ; begin write(chr(13)) end; { Carriage return }
procedure Bel; begin write(chr( 7)) end; { Beep }
procedure Cup; begin {PM} end;           { Cursor up }

function NotOk(question:alfa):boolean;
  var resp: char;
  function yn(c:char):boolean; begin yn:= c in ['y','n','Y','N'] end;
  begin
    repeat
      write(question,' ? (y/n) : '); ClrEol; resp:= Key;
      if not yn(resp) then begin Bel; Cr end
    until yn(resp);
    if resp in ['y','Y'] then write('yes') else write('no');
    NotOk:= resp in ['n','N']
  end;

procedure GetPolynm(var polynomial: PolynmTyp);
  var i, n : integer;

  function ok(n: integer): boolean;
    begin ok := (n >= 2) and (n <= C_MaxDegree) end;

  begin with polynomial do begin
    writeln;
    repeat
      write('Degree ? : '); ClrEol; readln(n);
      if not ok(n) then begin Bel; Cup end
    until ok(n); writeln; degree := n;
    writeln('Enter coefficients a[i] :');
    for i:= degree downto 0 do
      begin write('a[',i:C_NumDigits,'] = '); readln(coeff[i]) end;
    writeln;
    while (coeff[degree] = 0) and (degree > 0) do degree:= degree-1;
    curptr:= 0; denom:= 1;
  end end;

procedure WriteCoeff(a: integer; first, last: boolean);
  begin
    if a<0 then write(' - ') else if not first then write(' + ');
    a:= abs(a); if last or (a <> 1) then write(a)
  end;

procedure WriteExp(i: index); begin if i > 1 then write('^',i) end;

procedure WritePower(i: index);
begin if i>0 then begin write('x'); WriteExp(i) end end;

procedure WriteTerm(polynomial: PolynmTyp; i: index);
  var a: integer;
  begin with polynomial do begin
    WriteCoeff(coeff[i], i=degree, i=curptr);
    WritePower(i-curptr);
  end end;

procedure Display(polynomial: PolynmTyp);
  var i: index; parentheses: boolean;
  begin with polynomial do
    if degree=0 then WriteTerm(polynomial,0)
    else
      begin
        parentheses := (denom <> 1) or (curptr > 0);
        if parentheses then write(' (');
        for i:= degree downto curptr do
          if coeff[i] <> 0 then WriteTerm(polynomial, i);
        if parentheses then write(')')
      end
  end;

function GCD(a,b: integer): integer;
  var r: integer;
  begin
     if b<>0 then repeat r:= a mod b; a:= b; b:= r until r = 0;
     GCD:= a
  end;

procedure CommonFac(var polynomial: PolynmTyp);
{ Extract common factor }
  var i: index;
  begin with polynomial do begin
    curptr:= 0;
    while (curptr<degree) and (coeff[curptr] = 0) do curptr := curptr + 1;
    denom := abs(coeff[degree]);
    for i := degree-1 downto curptr do denom:= GCD(denom,abs(coeff[i]));
    if coeff[degree] < 0 then denom:= -denom;
    for i := curptr to degree do coeff[i] := coeff[i] div denom;
    if degree-curptr > 6 then writeln;
    WriteCoeff(denom,true,degree=0);
    WritePower(curptr)
{*};if debug then writeln
  end end;

function GetValue(polynomial: PolynmTyp; a: integer): integer;
  var y : integer; i: index;
  begin with polynomial do begin
    y := 0; for i:= degree downto curptr do y:= a*y + coeff[i];
    GetValue := y
{*};if debug then writeln('P(',a,') = ',y)
  end end;

procedure try(var polynomial: PolynmTyp; a: integer);
  var i, power: index;
  begin with polynomial do begin
    power := 0;
    while GetValue(polynomial, a) = 0 do
      begin
        power := power + 1;
        coeff[curptr] := a; curptr:= curptr+1;
        for i:= degree-1 downto curptr do coeff[i]:= coeff[i] + a * coeff[i+1];
      end;
    if power > 0 then
      begin
        write(' (x');
        if a < 0 then write('+'); write(-a);
        write(')'); WriteExp(power)
{*}    ;if debug then writeln
      end
  end end;

procedure Horner(var polynomial: PolynmTyp);
{ compute integer solutions }
  var a: integer;
  begin with polynomial do begin
    a:= 0;
    while a < abs(coeff[curptr]) do
      begin
        if a < abs(coeff[curptr] div 2) then a:= a+1 else a:= abs(coeff[curptr]);
        if (coeff[curptr] mod a) = 0 then
          begin
            try(polynomial, a);
            try(polynomial,-a)
          end
      end
  end end;

procedure Decompose(polynomial: PolynmTyp);
  begin with polynomial do begin
    Display(polynomial); write(' = ');
{*} if debug then writeln;
    CommonFac(polynomial);
    Horner(polynomial);
    if curptr < degree then Display(polynomial);
    writeln;
  end end;

begin
  repeat
    ClrScr;
    writeln('*** POLYNOMIAL DECOMPOSITION *** degree 2..',C_MaxDegree:C_NumDigits);
{*} if test then debug:= not NotOk('debug   ') else debug:= false;
    GetPolynm(polynomial);
    Decompose(polynomial);
    writeln
  until NotOk('Again   ')
end.
